import yaml


def test_introspect_write(cli, data_dir):
    r = cli(f'ansible-builder introspect {data_dir}')
    print(r.stdout)
    data = yaml.safe_load(r.stdout)  # assure that output is valid YAML
    assert 'python' in data
    assert 'system' in data
    assert 'pytz' in data['python']['test.reqfile']


def test_introspect_with_sanitize(cli, data_dir):
    r = cli(f'ansible-builder introspect --sanitize {data_dir}')
    print(r.stdout)
    data = yaml.safe_load(r.stdout)  # assure that output is valid YAML
    assert 'python' in data
    assert 'system' in data
    assert '# from collection test.bindep' in r.stdout  # should have comments



#!/usr/bin/env python3

import os
import sys
import yaml
import argparse


base_collections_path = '/usr/share/ansible/collections'
default_file = 'execution-environment.yml'


def line_is_empty(line):
    return bool((not line.strip()) or line.startswith('#'))


def pip_file_data(path):
    with open(path, 'r') as f:
        pip_content = f.read()

    pip_lines = []
    for line in pip_content.split('\n'):
        if line_is_empty(line):
            continue
        if line.startswith('-r') or line.startswith('--requirement'):
            _, new_filename = line.split(None, 1)
            new_path = os.path.join(os.path.dirname(path or '.'), new_filename)
            pip_lines.extend(pip_file_data(new_path))
        else:
            pip_lines.append(line)

    return pip_lines


def bindep_file_data(path):
    with open(path, 'r') as f:
        sys_content = f.read()

    sys_lines = []
    for line in sys_content.split('\n'):
        if line_is_empty(line):
            continue
        sys_lines.append(line)

    return sys_lines


def process_collection(path):
    """Return a tuple of (python_dependencies, system_dependencies) for the
    collection install path given.
    Both items returned are a list of dependencies.

    :param str path: root directory of collection (this would contain galaxy.yml file)
    """
    CD = CollectionDefinition(path)

    py_file = CD.get_dependency('python')
    pip_lines = []
    if py_file:
        pip_lines = pip_file_data(os.path.join(path, py_file))

    sys_file = CD.get_dependency('system')
    bindep_lines = []
    if sys_file:
        bindep_lines = bindep_file_data(os.path.join(path, sys_file))

    return (pip_lines, bindep_lines)


def process(data_dir=base_collections_path):
    paths = []
    path_root = os.path.join(data_dir, 'ansible_collections')

    # build a list of all the valid collection paths
    if os.path.exists(path_root):
        for namespace in sorted(os.listdir(path_root)):
            if not os.path.isdir(os.path.join(path_root, namespace)):
                continue
            for name in sorted(os.listdir(os.path.join(path_root, namespace))):
                collection_dir = os.path.join(path_root, namespace, name)
                if not os.path.isdir(collection_dir):
                    continue
                files_list = os.listdir(collection_dir)
                if 'galaxy.yml' in files_list or 'MANIFEST.json' in files_list:
                    paths.append(collection_dir)

    # populate the requirements content
    py_req = {}
    sys_req = {}
    for path in paths:
        col_pip_lines, col_sys_lines = process_collection(path)
        CD = CollectionDefinition(path)
        namespace, name = CD.namespace_name()
        key = '{}.{}'.format(namespace, name)

        if col_pip_lines:
            py_req[key] = col_pip_lines

        if col_sys_lines:
            sys_req[key] = col_sys_lines

    return {
        'python': py_req,
        'system': sys_req
    }


def has_content(candidate_file):
    """Beyond checking that the candidate exists, this also assures
    that the file has something other than whitespace,
    which can cause errors when given to pip.
    """
    if not os.path.exists(candidate_file):
        return False
    with open(candidate_file, 'r') as f:
        content = f.read()
    return bool(content.strip().strip('\n'))


class CollectionDefinition:
    """This class represents the dependency metadata for a collection
    should be replaced by logic to hit the Galaxy API if made available
    """

    def __init__(self, collection_path):
        self.reference_path = collection_path
        meta_file = os.path.join(collection_path, 'meta', default_file)
        if os.path.exists(meta_file):
            with open(meta_file, 'r') as f:
                self.raw = yaml.safe_load(f)
        else:
            self.raw = {'version': 1, 'dependencies': {}}
            # Automatically infer requirements for collection
            for entry, filename in [('python', 'requirements.txt'), ('system', 'bindep.txt')]:
                candidate_file = os.path.join(collection_path, filename)
                if has_content(candidate_file):
                    self.raw['dependencies'][entry] = filename

    def target_dir(self):
        namespace, name = self.namespace_name()
        return os.path.join(
            base_collections_path, 'ansible_collections',
            namespace, name
        )

    def namespace_name(self):
        "Returns 2-tuple of namespace and name"
        path_parts = [p for p in self.reference_path.split(os.path.sep) if p]
        return tuple(path_parts[-2:])

    def get_dependency(self, entry):
        """A collection is only allowed to reference a file by a relative path
        which is relative to the collection root
        """
        req_file = self.raw.get('dependencies', {}).get(entry)
        if req_file is None:
            return None
        elif os.path.isabs(req_file):
            raise RuntimeError(
                'Collections must specify relative paths for requirements files. '
                'The file {0} specified by {1} violates this.'.format(
                    req_file, self.reference_path
                )
            )

        return req_file


def simple_combine(reqs):
    """Given a dictionary of requirement lines keyed off collections,
    return a list with the most basic of de-duplication logic,
    and comments indicating the sources based off the collection keys
    """
    consolidated = []
    fancy_lines = []
    for collection, lines in reqs.items():
        for line in lines:
            if line_is_empty(line):
                continue

            base_line = line.split('#')[0].strip()
            if base_line in consolidated:
                i = consolidated.index(base_line)
                fancy_lines[i] += ', {}'.format(collection)
            else:
                fancy_line = base_line + '  # from collection {}'.format(collection)
                consolidated.append(base_line)
                fancy_lines.append(fancy_line)

    return fancy_lines


def add_introspect_options(parser):
    parser.add_argument(
        'folder', default=base_collections_path, nargs='?',
        help=(
            'Ansible collections path(s) to introspect. '
            'This should have a folder named ansible_collections inside of it.'
        )
    )


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        prog='ansible-builder-introspector',
        description=(
            'This is for programmatic use. '
            'Use ansible-builder introspect instead.'
        )
    )
    add_introspect_options(parser)
    args = parser.parse_args()
    data = process(args.folder)
    print(yaml.dump(data, default_flow_style=False))
    sys.exit(0)



import sys

from . import constants
from .exceptions import DefinitionError


class Steps:
    def __iter__(self):
        return iter(self.steps)


class AdditionalBuildSteps(Steps):
    def __init__(self, additional_steps):
        """Allows for additional prepended / appended build steps to be
        in the Containerfile or Dockerfile.
        """
        self.steps = []
        if isinstance(additional_steps, str):
            lines = additional_steps.strip().splitlines()
        elif isinstance(additional_steps, list):
            lines = additional_steps
        else:
            raise DefinitionError(
                """
                Error: Unknown type found for additional_build_steps; must be list or multi-line string.
                """
            )
            sys.exit(1)
        self.steps.extend(lines)

    def __iter__(self):
        return iter(self.steps)


class GalaxyInstallSteps(Steps):
    def __init__(self, requirements_naming):
        """Assumes given requirements file name has been placed in the build context
        """
        self.steps = []
        self.steps.append(
            "ADD {0} /build/{0}".format(requirements_naming)
        )
        self.steps.extend([
            "",
            "RUN ansible-galaxy role install -r /build/{0} --roles-path {1}".format(
                requirements_naming, constants.base_roles_path),
            "RUN ansible-galaxy collection install -r /build/{0} --collections-path {1}".format(
                requirements_naming, constants.base_collections_path),
            "",
            "RUN mkdir -p {0} {1}".format(constants.base_roles_path, constants.base_collections_path),
        ])


class GalaxyCopySteps(Steps):
    def __init__(self):
        """Assumes given requirements file name has been placed in the build context
        """
        self.steps = []
        self.steps.extend([
            "",
            "COPY --from=galaxy {0} {0}".format(constants.base_roles_path),
            "COPY --from=galaxy {0} {0}".format(constants.base_collections_path),
            "",
        ])


class AnsibleConfigSteps(Steps):
    def __init__(self, context_file):
        """Copies a user's ansible.cfg file for accessing Galaxy server"""
        self.steps = []
        self.steps.extend([
            f"ADD {context_file} ~/.ansible.cfg",
            "",
        ])



import os
import time
from pathlib import Path

import pytest

from ansible_builder.utils import write_file, copy_file


def test_write_file(tmpdir):
    path = os.path.join(tmpdir, 'foo.txt')
    text = [
        'foo  # from some collection',
        'bar',
        '# a comment',
        '',
        'zoo',
        ''  # trailing line
    ]
    assert write_file(path, text)  # does not exist, write
    assert not write_file(path, text)  # already correct, do not write


@pytest.fixture
def source_file(tmpdir):
    source = os.path.join(tmpdir, 'bar.txt')
    with open(source, 'w') as f:
        f.write('foo\nbar\n')
    return source


@pytest.fixture
def dest_file(tmpdir, source_file):
    '''Returns a file that has been copied from source file
    Use of fixture partially tests the copy_file functionality
    '''
    dest = os.path.join(tmpdir, 'foo.txt')
    assert copy_file(source_file, dest)
    assert not copy_file(source_file, dest)
    return dest


def test_copy_file(dest_file, source_file):
    # modify source file, which should trigger a re-copy
    with open(source_file, 'w') as f:
        f.write('foo\nbar\nzoo')

    assert copy_file(source_file, dest_file)
    assert not copy_file(source_file, dest_file)


def test_copy_touched_file(dest_file, source_file):
    # sleep for a miniscule amount of time, otherwise getmtime could be the same float value
    time.sleep(0.002)

    # touch does not change contents but updates modification time
    Path(source_file).touch()

    assert copy_file(source_file, dest_file)
    assert not copy_file(source_file, dest_file)



import os
import subprocess

import tempfile
import uuid

import pytest
import logging


logger = logging.getLogger(__name__)


TAG_PREFIX = 'quay.io/example/builder-test'


@pytest.fixture
def build_dir_and_ee_yml():
    """Fixture to return temporary file maker."""

    def tmp_dir_and_file(ee_contents):
        tmpdir = tempfile.mkdtemp(prefix="ansible-builder-test-")
        with tempfile.NamedTemporaryFile(delete=False, dir=tmpdir) as tempf:
            tempf.write(bytes(ee_contents, "UTF-8"))
        return tmpdir, tempf.name

    return tmp_dir_and_file


def run(args, *a, allow_error=False, **kw):
    kw["encoding"] = "utf-8"
    if "check" not in kw:
        # By default we want to fail if a command fails to run. Tests that
        # want to skip this can pass check=False when calling this fixture
        kw["check"] = True
    if "stdout" not in kw:
        kw["stdout"] = subprocess.PIPE
    if "stderr" not in kw:
        kw["stderr"] = subprocess.PIPE

    kw.setdefault("env", os.environ.copy()).update({"LANG": "en_US.UTF-8"})

    try:
        ret = CompletedProcessProxy(subprocess.run(args, shell=True, *a, **kw))
    except subprocess.CalledProcessError as err:
        if not allow_error:
            pytest.fail(
                f"Running {err.cmd} resulted in a non-zero return code: {err.returncode} - stdout: {err.stdout}, stderr: {err.stderr}"
            )
        err.rc = err.returncode  # lazyily make it look like a CompletedProcessProxy
        return err

    return ret


@pytest.fixture(scope='session', autouse=True)
def cleanup_ee_tags(container_runtime, request):
    def delete_images():
        r = run(f'{container_runtime} images --format="{{{{.Repository}}}}"')
        for image_name in r.stdout.split('\n'):
            from_test = False
            if not image_name:
                pass
            elif image_name.startswith('localhost/{0}'.format(TAG_PREFIX)):  # podman
                from_test = True
            elif image_name.startswith(TAG_PREFIX):  # docker
                from_test = True
            if from_test:
                run(f'{container_runtime} rmi -f {image_name}')
                logger.warning(f'Deleted image {image_name}')

    request.addfinalizer(delete_images)


@pytest.fixture()
def ee_tag(request, container_runtime):
    return '_'.join([
        TAG_PREFIX,
        request.node.name.lower().replace('[', '_').replace(']', '_'),
        str(uuid.uuid4())[:10]
    ])


@pytest.fixture(scope='class')
def ee_tag_class(request, container_runtime):
    return '_'.join([
        TAG_PREFIX,
        request.node.name.lower().replace('[', '_').replace(']', '_'),
        str(uuid.uuid4())[:10]
    ])


class CompletedProcessProxy(object):
    def __init__(self, result):
        self.result = result

    def __getattr__(self, attr):
        return getattr(self.result, attr)


@pytest.fixture()
def cli():
    return run


@pytest.fixture(scope='class')
def cli_class():
    return run


@pytest.fixture(params=['docker', 'podman'], ids=['docker', 'podman'], scope='session')
def container_runtime(request):
    if run(f'{request.param} --version', check=False).returncode == 0:
        return request.param
    else:
        pytest.skip(f'{request.param} runtime not available')



from ansible_builder.requirements import sanitize_requirements


def test_combine_entries():
    assert sanitize_requirements({
        'foo.bar': ['foo>1.0'],
        'bar.foo': ['foo>=2.0']
    }) == ['foo>1.0,>=2.0  # from collection foo.bar,bar.foo']


def test_remove_unwanted_requirements():
    assert sanitize_requirements({
        'foo.bar': [
            'foo',
            'ansible',
            'bar',
        ],
        'bar.foo': [
            'pytest',
            'bar',
            'zoo'
        ]
    }) == [
        'foo  # from collection foo.bar',
        'bar  # from collection foo.bar,bar.foo',
        'zoo  # from collection bar.foo'
    ]


def test_skip_bad_formats():
    """A single incorrectly formatted requirement should warn, but not block other reqs"""
    assert sanitize_requirements({'foo.bar': [
        'foo',
        'bar'
    ], 'foo.bad': ['zizzer zazzer zuzz']  # not okay
    }) == ['foo  # from collection foo.bar', 'bar  # from collection foo.bar']



import filecmp
import logging
import logging.config
import os
import shutil
import subprocess
import sys

from .colors import MessageColors


logger = logging.getLogger(__name__)
logging_levels = {
    '0': 'ERROR',
    '1': 'WARNING',
    '2': 'INFO',
    '3': 'DEBUG',
}


class ColorFilter(logging.Filter):
    color_map = {
        'ERROR': MessageColors.FAIL,
        'WARNING': MessageColors.WARNING,
        'INFO': MessageColors.HEADER,
        'DEBUG': MessageColors.OK
    }

    def filter(self, record):
        if sys.stdout.isatty():
            record.msg = self.color_map[record.levelname] + record.msg + MessageColors.ENDC
        return record


LOGGING = {
    'version': 1,
    'filters': {
        'colorize': {
            '()': ColorFilter
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'filters': ['colorize'],
            'stream': 'ext://sys.stdout'
        }
    },
    'loggers': {
        'ansible_builder': {
            'handlers': ['console'],
        }
    }
}


def configure_logger(verbosity):
    LOGGING['loggers']['ansible_builder']['level'] = logging_levels[str(verbosity)]
    logging.config.dictConfig(LOGGING)



def run_command(command, capture_output=False, allow_error=False):
    logger.info('Running command:')
    logger.info('  {0}'.format(' '.join(command)))
    try:
        process = subprocess.Popen(command,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT)
    except FileNotFoundError:
        logger.error(f"You do not have {command[0]} installed, please specify a different container runtime for this command.")
        sys.exit(1)

    output = []
    for line in iter(process.stdout.readline, b''):
        line = line.decode(sys.stdout.encoding)
        if capture_output:
            output.append(line.rstrip())
        logger.debug(line.rstrip('\n'))  # line ends added by logger itself
    logger.debug('')

    rc = process.poll()
    if rc is not None and rc != 0 and (not allow_error):
        main_logger = logging.getLogger('ansible_builder')
        if main_logger.level > logging.INFO:
            logger.error('Command that had error:')
            logger.error('  {0}'.format(' '.join(command)))
        if main_logger.level > logging.DEBUG:
            for line in output:
                logger.error(line)
            logger.error('')
        logger.error(f"An error occured (rc={rc}), see output line(s) above for details.")
        sys.exit(1)

    return (rc, output)


def write_file(filename: str, lines: list) -> bool:
    new_text = '\n'.join(lines)
    if os.path.exists(filename):
        with open(filename, 'r') as f:
            if f.read() == new_text:
                logger.debug("File {0} is already up-to-date.".format(filename))
                return False
            else:
                logger.warning('File {0} had modifications and will be rewritten'.format(filename))
    with open(filename, 'w') as f:
        f.write(new_text)
    return True


def copy_file(source: str, dest: str) -> bool:
    should_copy = False

    if not os.path.exists(dest):
        logger.debug("File {0} will be created.".format(dest))
        should_copy = True
    elif not filecmp.cmp(source, dest, shallow=False):
        logger.warning('File {0} had modifications and will be rewritten'.format(dest))
        should_copy = True
    elif os.path.getmtime(source) > os.path.getmtime(dest):
        logger.warning('File {0} updated time increased and will be rewritten'.format(dest))
        should_copy = True

    if should_copy:
        shutil.copy(source, dest)
    else:
        logger.debug("File {0} is already up-to-date.".format(dest))

    return should_copy



import logging
import os
import textwrap
import yaml

from . import constants
from .exceptions import DefinitionError
from .steps import (
    AdditionalBuildSteps, GalaxyInstallSteps, GalaxyCopySteps, AnsibleConfigSteps
)
from .utils import run_command, write_file, copy_file
from .requirements import sanitize_requirements
import ansible_builder.introspect


logger = logging.getLogger(__name__)

# Files that need to be moved into the build context, and their naming inside the context
CONTEXT_FILES = {
    'galaxy': 'requirements.yml'
}

# Subdirectory in the build context where files to ADD get copied
CONTEXT_BUILD_OUTPUTS_DIR = '_build'

BINDEP_COMBINED = 'bindep_combined.txt'
PIP_COMBINED = 'requirements_combined.txt'

ALLOWED_KEYS = [
    'version',
    'base_image',
    'dependencies',
    'ansible_config',
    'additional_build_steps',
]


class AnsibleBuilder:
    def __init__(self, action=None,
                 filename=constants.default_file,
                 base_image=None,
                 build_context=constants.default_build_context,
                 tag=constants.default_tag,
                 container_runtime=constants.default_container_runtime,
                 verbosity=2):
        self.action = action
        self.definition = UserDefinition(filename=filename)

        # Handle precedence of the base image
        if base_image is not None:
            self.base_image = base_image
        if base_image is None:
            if self.definition.raw.get('base_image'):
                self.base_image = self.definition.raw.get('base_image')
            else:
                self.base_image = constants.default_base_image

        self.tag = tag
        self.build_context = build_context
        self.build_outputs_dir = os.path.join(build_context, CONTEXT_BUILD_OUTPUTS_DIR)
        self.container_runtime = container_runtime
        self.containerfile = Containerfile(
            definition=self.definition,
            base_image=self.base_image,
            build_context=self.build_context,
            container_runtime=self.container_runtime,
            tag=self.tag)
        self.verbosity = verbosity

    @property
    def version(self):
        return self.definition.version

    @property
    def ansible_config(self):
        return self.definition.ansible_config

    @property
    def build_command(self):
        return [
            self.container_runtime, "build",
            "-f", self.containerfile.path,
            "-t", self.tag,
            self.build_context
        ]

    def run_in_container(self, command, **kwargs):
        wrapped_command = [self.container_runtime, 'run', '--rm']

        wrapped_command.extend(['-v', f"{os.path.abspath(self.build_context)}:/context:Z"])

        wrapped_command.extend([self.tag] + command)

        return run_command(wrapped_command, **kwargs)

    def run_intermission(self):
        run_command(self.build_command, capture_output=True)

        in_container_introspect_path = os.path.join(
            '/context', CONTEXT_BUILD_OUTPUTS_DIR, 'introspect.py'
        )
        rc, introspect_output = self.run_in_container(
            ['python3', in_container_introspect_path], capture_output=True
        )
        collection_data = yaml.safe_load('\n'.join(introspect_output))

        # Add data from user definition, go from dicts to list
        collection_data['system']['user'] = self.definition.user_system
        collection_data['python']['user'] = self.definition.user_python
        system_lines = ansible_builder.introspect.simple_combine(collection_data['system'])
        python_lines = sanitize_requirements(collection_data['python'])

        if system_lines:
            bindep_file = os.path.join(self.build_outputs_dir, BINDEP_COMBINED)
            write_file(bindep_file, system_lines + [''])

        if python_lines:
            pip_file = os.path.join(self.build_outputs_dir, PIP_COMBINED)
            write_file(pip_file, python_lines)

        return (system_lines, python_lines)

    def build(self):
        # Phase 1 of Containerfile
        self.containerfile.create_folder_copy_files()
        self.containerfile.prepare_ansible_config_file()
        self.containerfile.prepare_galaxy_install_steps()
        logger.debug('Writing partial Containerfile without collection requirements')
        self.containerfile.write()
        system_lines, python_lines = self.run_intermission()

        # Phase 2 of Containerfile
        self.containerfile.prepare_build_stage_steps()
        self.containerfile.prepare_assemble_steps()

        self.containerfile.prepare_final_stage_steps()
        self.containerfile.prepare_prepended_steps()
        self.containerfile.prepare_galaxy_copy_steps()
        self.containerfile.prepare_system_runtime_deps_steps()
        self.containerfile.prepare_appended_steps()
        logger.debug('Rewriting Containerfile to capture collection requirements')
        self.containerfile.write()
        run_command(self.build_command)
        return True


class BaseDefinition:
    """Subclasses should populate these properties in the __init__ method
    self.raw - a dict that basically is the definition
    self.reference_path - the folder which dependencies are specified relative to
    """

    @property
    def version(self):
        version = self.raw.get('version')

        if not version:
            raise ValueError("Expected top-level 'version' key to be present.")

        return str(version)

    @property
    def ansible_config(self):
        ansible_config = self.raw.get('ansible_config')

        if not ansible_config:
            pass
        else:
            return str(ansible_config)


class UserDefinition(BaseDefinition):
    def __init__(self, filename):
        self.filename = filename
        self.reference_path = os.path.dirname(filename)

        try:
            with open(filename, 'r') as f:
                y = yaml.safe_load(f)
                self.raw = y if y else {}
        except FileNotFoundError:
            raise DefinitionError(textwrap.dedent("""
            Could not detect '{0}' file in this directory.
            Use -f to specify a different location.
            """).format(filename))
        except (yaml.parser.ParserError, yaml.scanner.ScannerError) as e:
            raise DefinitionError(textwrap.dedent("""
            An error occured while parsing the definition file:
            {0}
            """).format(str(e)))

        if not isinstance(self.raw, dict):
            raise DefinitionError("Definition must be a dictionary, not {0}".format(type(self.raw).__name__))

        self.user_python = self.read_dependency('python')
        self.user_system = self.read_dependency('system')

    def get_additional_commands(self):
        """Gets additional commands from the exec env file, if any are specified.
        """
        commands = self.raw.get('additional_build_steps')
        return commands

    def get_dep_abs_path(self, entry):
        """Unique to the user EE definition, files can be referenced by either
        an absolute path or a path relative to the EE definition folder
        This method will return the absolute path.
        """
        req_file = self.raw.get('dependencies', {}).get(entry)

        if not req_file:
            return None

        if os.path.isabs(req_file):
            return req_file

        return os.path.join(self.reference_path, req_file)

    def read_dependency(self, entry):
        requirement_path = self.get_dep_abs_path(entry)
        if not requirement_path:
            return []
        try:
            with open(requirement_path, 'r') as f:
                return f.read().split('\n')
        except FileNotFoundError:
            raise DefinitionError("Dependency file {0} does not exist.".format(requirement_path))

    def validate(self):
        # Check that all specified keys in the definition file are valid.
        def_file_dict = self.raw
        yaml_keys = set(def_file_dict.keys())
        invalid_keys = yaml_keys - set(ALLOWED_KEYS)
        if invalid_keys:
            raise DefinitionError(textwrap.dedent(
                f"""
                Error: Unknown yaml key(s), {invalid_keys}, found in the definition file.\n
                Allowed options are:
                {ALLOWED_KEYS}
                """)
            )

        for item in CONTEXT_FILES:
            requirement_path = self.get_dep_abs_path(item)
            if requirement_path:
                if not os.path.exists(requirement_path):
                    raise DefinitionError("Dependency file {0} does not exist.".format(requirement_path))

        ee_base_image = self.raw.get('base_image')
        if ee_base_image:
            if not isinstance(ee_base_image, str):
                raise DefinitionError(textwrap.dedent(
                    f"""
                    Error: Unknown type {type(ee_base_image)} found for base_image; must be a string.
                    """)
                )

        additional_cmds = self.get_additional_commands()
        if additional_cmds:
            if not isinstance(additional_cmds, dict):
                raise DefinitionError(textwrap.dedent("""
                    Expected 'additional_build_steps' in the provided definition file to be a dictionary
                    with keys 'prepend' and/or 'append'; found a {0} instead.
                    """).format(type(additional_cmds).__name__))

            expected_keys = frozenset(('append', 'prepend'))
            unexpected_keys = set(additional_cmds.keys()) - expected_keys
            if unexpected_keys:
                raise DefinitionError(
                    f"Keys {*unexpected_keys,} are not allowed in 'additional_build_steps'."
                )

        ansible_config_path = self.raw.get('ansible_config')
        if ansible_config_path:
            if not isinstance(ansible_config_path, str):
                raise DefinitionError(textwrap.dedent("""
                    Expected 'ansible_config' in the provided definition file to
                    be a string; found a {0} instead.
                    """).format(type(ansible_config_path).__name__))


class Containerfile:
    newline_char = '\n'

    def __init__(self, definition,
                 build_context=None,
                 base_image=None,
                 container_runtime=None,
                 tag=None):

        self.build_context = build_context
        self.build_outputs_dir = os.path.join(build_context, CONTEXT_BUILD_OUTPUTS_DIR)
        self.definition = definition
        filename = constants.runtime_files[container_runtime]
        self.path = os.path.join(self.build_context, filename)
        self.base_image = base_image
        self.builder_stage_image = 'quay.io/ansible/python-builder:latest'
        self.container_runtime = container_runtime
        self.tag = tag
        self.steps = [
            "FROM {0} as galaxy".format(self.base_image),
            ""
        ]

    def create_folder_copy_files(self):
        """Creates the build context file for this Containerfile
        moves files from the definition into the folder
        """
        # courteously validate items before starting to write files
        self.definition.validate()

        os.makedirs(self.build_outputs_dir, exist_ok=True)

        for item, new_name in CONTEXT_FILES.items():
            requirement_path = self.definition.get_dep_abs_path(item)
            if requirement_path is None:
                continue
            dest = os.path.join(self.build_context, CONTEXT_BUILD_OUTPUTS_DIR, new_name)
            copy_file(requirement_path, dest)

        # copy introspect.py file from source into build context
        copy_file(
            ansible_builder.introspect.__file__,
            os.path.join(self.build_outputs_dir, 'introspect.py')
        )

        if self.definition.ansible_config:
            copy_file(
                self.definition.ansible_config,
                os.path.join(self.build_outputs_dir, 'ansible.cfg')
            )

    def prepare_ansible_config_file(self):
        ansible_config_file_path = self.definition.ansible_config
        if ansible_config_file_path:
            context_file_path = os.path.join(CONTEXT_BUILD_OUTPUTS_DIR, 'ansible.cfg')
            return self.steps.extend(AnsibleConfigSteps(context_file_path))

    def prepare_prepended_steps(self):
        additional_prepend_steps = self.definition.get_additional_commands()
        if additional_prepend_steps:
            prepended_steps = additional_prepend_steps.get('prepend')
            if prepended_steps:
                return self.steps.extend(AdditionalBuildSteps(prepended_steps))

        return False

    def prepare_appended_steps(self):
        additional_append_steps = self.definition.get_additional_commands()
        if additional_append_steps:
            appended_steps = additional_append_steps.get('append')
            if appended_steps:
                return self.steps.extend(AdditionalBuildSteps(appended_steps))

        return False

    def prepare_galaxy_install_steps(self):
        if self.definition.get_dep_abs_path('galaxy'):
            relative_galaxy_path = os.path.join(CONTEXT_BUILD_OUTPUTS_DIR, CONTEXT_FILES['galaxy'])
            self.steps.extend(GalaxyInstallSteps(relative_galaxy_path))
        return self.steps

    def prepare_assemble_steps(self):
        requirements_file_exists = os.path.exists(os.path.join(self.build_outputs_dir, PIP_COMBINED))
        if requirements_file_exists:
            relative_requirements_path = os.path.join(CONTEXT_BUILD_OUTPUTS_DIR, PIP_COMBINED)
            self.steps.append(f"ADD {relative_requirements_path} /tmp/src/requirements.txt")

        bindep_exists = os.path.exists(os.path.join(self.build_outputs_dir, BINDEP_COMBINED))
        if bindep_exists:
            relative_bindep_path = os.path.join(CONTEXT_BUILD_OUTPUTS_DIR, BINDEP_COMBINED)
            self.steps.append(f"ADD {relative_bindep_path} /tmp/src/bindep.txt")

        if requirements_file_exists or bindep_exists:
            self.steps.append("RUN assemble")

        return self.steps

    def prepare_system_runtime_deps_steps(self):
        requirements_file_exists = os.path.exists(os.path.join(self.build_outputs_dir, PIP_COMBINED))
        bindep_exists = os.path.exists(os.path.join(self.build_outputs_dir, BINDEP_COMBINED))

        if requirements_file_exists or bindep_exists:
            self.steps.extend([
                "COPY --from=builder /output/ /output/",
                "RUN /output/install-from-bindep && rm -rf /output/wheels",
            ])

        return self.steps

    def prepare_build_stage_steps(self):
        self.steps.extend([
            "",
            "FROM {0} as builder".format(self.builder_stage_image),
            "",
        ])
        return self.steps

    def prepare_final_stage_steps(self):
        self.steps.extend([
            "",
            "FROM {0}".format(self.base_image),
            "",
        ])
        return self.steps

    def prepare_galaxy_copy_steps(self):
        if self.definition.get_dep_abs_path('galaxy'):
            self.steps.extend(GalaxyCopySteps())
        return self.steps

    def write(self):
        with open(self.path, 'w') as f:
            for step in self.steps:
                f.write(step + self.newline_char)

        return True



import re


def test_help(cli):
    result = cli('ansible-builder --help', check=False)
    help_text = result.stdout
    assert 'usage: ansible-builder [-h] [--version] {build,introspect}' in help_text


def test_no_args(cli):
    result = cli('ansible-builder', check=False)
    stderr = result.stderr
    assert 'usage: ansible-builder [-h] [--version] {build,introspect}' in stderr
    assert 'ansible-builder: error: the following arguments are required: action' in stderr


def test_version(cli):
    result = cli('ansible-builder --version')
    version = result.stdout
    matches = re.findall(r'\d.\d.\d', version)
    assert matches



import argparse
import logging
import sys
import yaml

from . import __version__

from .colors import MessageColors
from .exceptions import DefinitionError
from .main import AnsibleBuilder
from . import constants
from .introspect import add_introspect_options, process, simple_combine
from .requirements import sanitize_requirements
from .utils import configure_logger


logger = logging.getLogger(__name__)


def run():
    args = parse_args()
    configure_logger(args.verbosity)

    if args.action in ['build']:
        logger.debug(f'Ansible Builder is building your execution environment image, "{args.tag}".')
        ab = AnsibleBuilder(**vars(args))
        action = getattr(ab, ab.action)
        try:
            if action():
                print(MessageColors.OKGREEN + "Complete! The build context can be found at: {0}".format(ab.build_context) + MessageColors.ENDC)
                sys.exit(0)
        except DefinitionError as e:
            logger.error(e.args[0])
            sys.exit(1)
    elif args.action == 'introspect':
        data = process(args.folder)
        if args.sanitize:
            data['python'] = sanitize_requirements(data['python'])
            data['system'] = simple_combine(data['system'])
            logger.info('# Sanitized dependencies for {0}'.format(args.folder))
        else:
            print('# Dependency data for {0}'.format(args.folder))
        print('---')
        print(yaml.dump(data, default_flow_style=False))
        sys.exit(0)

    logger.error("An error has occured.")
    sys.exit(1)


def parse_args(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(
        prog='ansible-builder',
        description=(
            'Tooling to help build container images for running Ansible content. '
            'Get started by looking at the help text for one of the subcommands.'
        )
    )
    parser.add_argument(
        '--version', action='version', version=__version__,
        help='Print ansible-builder version and exit.'
    )
    subparsers = parser.add_subparsers(help='The command to invoke.', dest='action')
    subparsers.required = True # This can be a kwarg in python 3.7+

    build_command_parser = subparsers.add_parser(
        'build',
        help='Builds a container image.',
        description=(
            'Creates a build context (including a Containerfile) from an execution environment spec. '
            'The build context will be populated from the execution environment spec. '
            'After that, the specified container runtime podman/docker will be invoked to '
            'build an image from that definition. '
            'After building the image, it can be used locally or published using the supplied tag.'
        )
    )

    build_command_parser.add_argument('-t', '--tag',
                                      default=constants.default_tag,
                                      help='The name for the container image being built.')

    for p in [build_command_parser]:

        p.add_argument('-f', '--file',
                       default=constants.default_file,
                       dest='filename',
                       help='The definition of the execution environment.')

        p.add_argument('-b', '--base-image',
                       default=None,
                       help='The parent image for the execution environment.')

        p.add_argument('-c', '--context',
                       default=constants.default_build_context,
                       dest='build_context',
                       help='The directory to use for the build context. Defaults to $PWD/context.')

        p.add_argument('--container-runtime',
                       choices=list(constants.runtime_files.keys()),
                       default=constants.default_container_runtime,
                       help='Specifies which container runtime to use. Defaults to {0}.'.format(constants.default_container_runtime))

        p.add_argument('-v', '--verbosity',
                       dest='verbosity',
                       type=int,
                       choices=[0, 1, 2, 3],
                       default=2,
                       help='Increase the output verbosity, for up to three levels of verbosity '
                            '(invoked via "--verbosity" or "-v" followed by an integer ranging '
                            'in value from 0 to 3)')

    introspect_parser = subparsers.add_parser(
        'introspect',
        help='Introspects collections in folder.',
        description=(
            'Loops over collections in folder and returns data about dependencies. '
            'This is used internally and exposed here for verification. '
            'This is targeted toward collection authors and maintainers.'
        )
    )
    add_introspect_options(introspect_parser)
    introspect_parser.add_argument(
        '--sanitize', help=(
            'Sanitize and de-duplicate requirements. '
            'This is normally done separately from the introspect script, but this '
            'option is given to more accurately test collection content.'
        ), action='store_true')
    introspect_parser.add_argument(
        '-v', '--verbosity', dest='verbosity', action='count', default=0, help=(
            'Increase the output verbosity, for up to three levels of verbosity '
            '(invoked via "--verbosity" or "-v" followed by an integer ranging '
            'in value from 0 to 3)'))

    args = parser.parse_args(args)

    return args



from ansible_builder.main import AnsibleBuilder
from ansible_builder.cli import parse_args


def prepare(args):
    args = parse_args(args)
    return AnsibleBuilder(**vars(args))


def test_custom_image(exec_env_definition_file, tmpdir):
    content = {'version': 1}
    path = str(exec_env_definition_file(content=content))

    aee = prepare(['build', '-f', path, '-b', 'my-custom-image', '-c', str(tmpdir)])

    assert aee.containerfile.base_image == 'my-custom-image'


def test_build_context(good_exec_env_definition_path, tmpdir):
    path = str(good_exec_env_definition_path)
    build_context = str(tmpdir)
    aee = prepare(['build', '-f', path, '-c', build_context])

    assert aee.build_context == build_context



import pytest
import os


def test_definition_syntax_error(cli, data_dir):
    ee_def = os.path.join(data_dir, 'definition_files', 'invalid.yml')
    r = cli(
        f'ansible-builder build -f {ee_def} --container-runtime podman',
        allow_error=True
    )
    assert r.rc != 0
    assert 'An error occured while parsing the definition file' in (r.stdout + r.stderr), (r.stdout + r.stderr)


def test_build_fail_exitcode(cli, container_runtime, ee_tag, tmpdir, data_dir):
    """Test that when a build fails, the ansible-builder exits with non-zero exit code.

    Example: https://github.com/ansible/ansible-builder/issues/51
    """
    bc = str(tmpdir)
    ee_def = os.path.join(data_dir, 'build_fail', 'execution-environment.yml')
    r = cli(
        f"ansible-builder build -c {bc} -f {ee_def} -t {ee_tag} --container-runtime {container_runtime} -v3",
        allow_error=True
    )
    assert r.rc != 0
    assert 'RUN thisisnotacommand' in (r.stdout + r.stderr), (r.stdout + r.stderr)
    assert 'thisisnotacommand: command not found' in (r.stdout + r.stderr), (r.stdout + r.stderr)


def test_missing_python_requirements_file():
    """If a user specifies a python requirements file, but we can't find it, fail sanely."""
    pytest.skip("Not implemented")


def test_missing_galaxy_requirements_file():
    """If a user specifies a galaxy requirements file, but we can't find it, fail sanely."""
    pytest.skip("Not implemented")


def test_build_streams_output_with_verbosity_on(cli, container_runtime, build_dir_and_ee_yml, ee_tag):
    """Test that 'ansible-builder build' streams build output."""
    tmpdir, eeyml = build_dir_and_ee_yml("")
    result = cli(f"ansible-builder build -c {tmpdir} -f {eeyml} -t {ee_tag} --container-runtime {container_runtime} -v 3")
    assert f'{container_runtime} build -f {tmpdir}' in result.stdout
    assert f'Ansible Builder is building your execution environment image, "{ee_tag}".' in result.stdout
    assert f'The build context can be found at: {tmpdir}' in result.stdout


def test_build_streams_output_with_verbosity_off(cli, container_runtime, build_dir_and_ee_yml, ee_tag):
    """
    Like the test_build_streams_output_with_verbosity_on test but making sure less output is shown with default verbosity level of 2.
    """
    tmpdir, eeyml = build_dir_and_ee_yml("")
    result = cli(f"ansible-builder build -c {tmpdir} -f {eeyml} -t {ee_tag} --container-runtime {container_runtime}")
    assert f'Ansible Builder is building your execution environment image, "{ee_tag}".' not in result.stdout
    assert f'The build context can be found at: {tmpdir}' in result.stdout


def test_build_streams_output_with_invalid_verbosity(cli, container_runtime, build_dir_and_ee_yml, ee_tag):
    """
    Like the test_build_streams_output_with_verbosity_off test but making sure it errors out correctly with invalid verbosity level.
    """
    tmpdir, eeyml = build_dir_and_ee_yml("")
    result = cli(f"ansible-builder build -c {tmpdir} -f {eeyml} -t {ee_tag} --container-runtime {container_runtime} -v 6", allow_error=True)
    assert result.rc != 0
    assert 'invalid choice: 6 (choose from 0, 1, 2, 3)' in (result.stdout + result.stderr)


def test_blank_execution_environment(cli, container_runtime, ee_tag, tmpdir, data_dir):
    """Just makes sure that the buld process does not require any particular input"""
    bc = str(tmpdir)
    ee_def = os.path.join(data_dir, 'blank', 'execution-environment.yml')
    cli(
        f'ansible-builder build -c {bc} -f {ee_def} -t {ee_tag} --container-runtime {container_runtime}'
    )
    result = cli(f'{container_runtime} run --rm {ee_tag} echo "This is a simple test"')
    assert 'This is a simple test' in result.stdout, result.stdout


def test_user_system_requirement(cli, container_runtime, ee_tag, tmpdir, data_dir):
    bc = str(tmpdir)
    ee_def = os.path.join(data_dir, 'subversion', 'execution-environment.yml')
    command = f'ansible-builder build -c {bc} -f {ee_def} -t {ee_tag} --container-runtime {container_runtime}'
    cli(command)
    result = cli(
        f'{container_runtime} run --rm {ee_tag} svn --help'
    )
    assert 'Subversion is a tool for version control' in result.stdout, result.stdout


def test_collection_system_requirement(cli, container_runtime, ee_tag, tmpdir, data_dir):
    bc = str(tmpdir)
    ee_def = os.path.join(data_dir, 'ansible.posix.at', 'execution-environment.yml')
    cli(
        f'ansible-builder build -c {bc} -f {ee_def} -t {ee_tag} --container-runtime {container_runtime} -v3'
    )
    result = cli(
        f'{container_runtime} run --rm {ee_tag} at -V'
    )
    assert 'at version' in result.stderr, result.stderr


def test_user_python_requirement(cli, container_runtime, ee_tag, tmpdir, data_dir):
    bc = str(tmpdir)
    ee_def = os.path.join(data_dir, 'pip', 'execution-environment.yml')
    command = f'ansible-builder build -c {bc} -f {ee_def} -t {ee_tag} --container-runtime {container_runtime}'
    cli(command)
    result = cli(
        f'{container_runtime} run --rm {ee_tag} pip3 show awxkit'
    )
    assert 'The official command line interface for Ansible AWX' in result.stdout, result.stdout


def test_prepended_steps(cli, container_runtime, ee_tag, tmpdir, data_dir):
    """
    Tests that prepended steps are in final stage
    """
    bc = str(tmpdir)
    ee_def = os.path.join(data_dir, 'prepend_steps', 'execution-environment.yml')
    cli(
        f'ansible-builder build -c {bc} -f {ee_def} -t {ee_tag} --container-runtime {container_runtime}'
    )

    _file = 'Dockerfile' if container_runtime == 'docker' else 'Containerfile'
    content = open(os.path.join(bc, _file), 'r').read()

    stages_content = content.split('FROM')

    assert 'RUN whoami' in stages_content[-1]


class TestPytz:

    @pytest.fixture(scope='class')
    def pytz(self, cli_class, container_runtime, ee_tag_class, data_dir, tmpdir_factory):
        bc_folder = str(tmpdir_factory.mktemp('bc'))
        ee_def = os.path.join(data_dir, 'pytz', 'execution-environment.yml')
        r = cli_class(
            f'ansible-builder build -c {bc_folder} -f {ee_def} -t {ee_tag_class} --container-runtime {container_runtime} -v 3'
        )
        assert 'Collecting pytz' in r.stdout, r.stdout
        return (ee_tag_class, bc_folder)

    def test_has_pytz(self, cli, container_runtime, pytz):
        ee_tag, bc_folder = pytz
        r = cli(f'{container_runtime} run --rm {ee_tag} pip3 show pytz')
        assert 'World timezone definitions, modern and historical' in r.stdout, r.stdout

    def test_build_layer_reuse(self, cli, container_runtime, data_dir, pytz):
        ee_tag, bc_folder = pytz
        ee_def = os.path.join(data_dir, 'pytz', 'execution-environment.yml')
        r = cli(
            f'ansible-builder build -c {bc_folder} -f {ee_def} -t {ee_tag} --container-runtime {container_runtime} -v 3'
        )
        assert 'Collecting pytz (from -r /build/' not in r.stdout, r.stdout
        assert 'requirements_combined.txt is already up-to-date' in r.stdout, r.stdout
        stdout_no_whitespace = r.stdout.replace('--->', '-->').replace('\n', ' ').replace('   ', ' ').replace('  ', ' ')
        assert 'RUN /output/install-from-bindep --> Using cache' in stdout_no_whitespace, r.stdout



import yaml
from unittest import mock
import pytest
import os


@pytest.fixture(autouse=True)
def do_not_run_commands():
    cmd_mock = mock.MagicMock(return_value=[1, [
        'python:', '  foo: []', 'system: {}',
    ]])
    with mock.patch('ansible_builder.main.run_command', new=cmd_mock):
        yield cmd_mock


@pytest.fixture(scope='session')
def data_dir():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), 'data'))


@pytest.fixture
def exec_env_definition_file(tmpdir):

    def _write_file(content=None):
        path = tmpdir.mkdir('aee').join('execution-env.yml')

        write_str = {}
        if isinstance(content, dict):
            write_str = yaml.dump(content)
        elif isinstance(content, str):
            write_str = content

        with open(path, 'w') as outfile:
            outfile.write(write_str)

        return path

    return _write_file


good_content = {'version': 1}


@pytest.fixture
def good_exec_env_definition_path(tmpdir):
    path = tmpdir.mkdir('aee').join('execution-env.yml')

    with open(path, 'w') as outfile:
        yaml.dump(good_content, outfile)

    return str(path)


@pytest.fixture
def galaxy_requirements_file(tmpdir):

    def _write_file(content={}):
        path = tmpdir.mkdir('galaxy').join('requirements.yml')

        with open(path, 'w') as outfile:
            yaml.dump(content, outfile)

        return path

    return _write_file
